pub mod layout;

use cdtypes::types::cdstring::DString;
use tool_helper::Error;

pub struct CDDesc {
    system_area: SystemArea,
    pvd:         PrimaryVolumeDescriptor,
    pub root:    Directory
}

impl CDDesc {
    pub fn new() -> CDDesc {
        match Directory::new("root") {
            Ok(root)   => CDDesc{system_area: SystemArea::new(), pvd: PrimaryVolumeDescriptor::new(), root},
            Err(error) => panic!("Creating root directory failed with: {}", error)
        }
    }

    pub fn get_memory_layout(&self) -> layout::MemoryLayout {
        layout::DefaultLayout::new(self)
    }

    pub fn get_memory_layout_mut(&mut self) -> layout::MemoryLayoutMut {
        layout::DefaultLayoutMut::new(self)
    }
}

pub struct SystemArea {
}

impl SystemArea {
    pub fn new() -> SystemArea {
        SystemArea{}
    }
}

pub struct PrimaryVolumeDescriptor {
}

impl PrimaryVolumeDescriptor {
    pub fn new() -> PrimaryVolumeDescriptor {
        PrimaryVolumeDescriptor{}
    }
}

pub struct Directory {
    name:       DirectoryName,
    properties: Properties,
    data:       Vec<Data>,
    dirs:       Vec<Directory>
}

impl Directory {
    pub fn new(dir_name: &str) -> Result<Directory, Error> {
        Ok(Directory{name: DirectoryName::from_str(dir_name)?, properties: Properties::default(), data: Vec::new(), dirs: Vec::new()})
    }

    pub fn add_dir(&mut self, dir: Directory) {
        self.dirs.push(dir);
    }

    pub fn add_data(&mut self, data: Data) {
        self.data.push(data);
    }
}

impl std::fmt::Display for Directory {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "Directory: {}", self.name)
    }    
}

pub struct Data {
    name:       FileName,
    properties: Properties
}

impl Data {
    pub fn new(file_name: &str) -> Result<Data, Error> {
        Ok(Data{name: FileName::from_str(file_name)?, properties: Properties::default()})
    }
}

impl std::fmt::Display for Data {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "File: {}", self.name)
    }
}

pub struct DirectoryName {
    name: DString<8>,
}

impl DirectoryName {
    pub fn from_str(dir_name: &str) -> Result<DirectoryName, Error> {
        let dir_name = dir_name.to_uppercase();
        Ok(DirectoryName{name: DString::from_str(dir_name.as_ref())?})
    }
}

impl std::fmt::Display for DirectoryName {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", dstring_as_str(&self.name))
    } 
}

pub struct FileName {
    name: DString<8>,
    ext:  Option<DString<3>>
}

impl FileName {
    pub fn from_str(file_name: &str) -> Result<FileName, Error> {
        let file_name = file_name.to_uppercase();
        let (name, ext) = {
            let mut sub_str = file_name.split('.');
            (Error::ok_or_new(sub_str.next(), || "File name can't be emplty".to_owned())?, sub_str.next())
        };

        Ok(FileName{name: DString::from_str(name)?, ext: {
            match ext {
                Some(ext) => Some(DString::from_str(ext)?),
                None      => None
            }
        }})
    }
}

impl std::fmt::Display for FileName {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        if let Some(ext) = &self.ext {
            write!(f, "{}.{};1", dstring_as_str(&self.name), dstring_as_str(&ext))
        }

        else {
            write!(f, "{}", dstring_as_str(&self.name))
        }
    }
}

pub struct Properties {
    pub lba:    Option<usize>,
    pub size_b: Option<usize>
}

impl Default for Properties {
    fn default() -> Self {
        Properties{lba: None, size_b: None}
    }
}

fn dstring_as_str<const SIZE: usize>(string: &DString<SIZE>) -> &str {
    match std::str::from_utf8(string.as_raw()) {
        Ok(str) => str,
        Err(_)  => "???",
    }
}